@@ -3,6 +3,8 @@
 import logging
 from enum import IntEnum, unique
 
+import numpy as np
+import numpy.typing as npt
 import pandas as pd
 
 
@@ -44,28 +46,9 @@ class ProbabilityVolumeResults(IntEnum):
     MV_DIS2_ACT = 16
 
 
-def export_results(result_data, outfile, run_config):
+def export_results(result_data: dict[str, npt.NDArray[np.float64]], outfile: str) -> None:
+    """ Write the results dict :result_data to the output file :param: """
 
     logger.info("Exporting NCF to %s", outfile)
-
-    # time_axis = projector.time_axis
-
-    # data = {
-    #     "YEAR": time_axis.years,
-    #     "QUARTER": time_axis.quarters,
-    #     "MONTH": time_axis.months,
-    #     "PREMIUM": projector.ncf_portfolio[:, 0],
-    #     "DI_ONG_CLAIMS1": projector.ncf_portfolio[:, 1],
-    #     "DI_ONG_CLAIMS2": projector.ncf_portfolio[:, 2]
-    # }
-
-    # # add the reserves
-    # for st in projector.model.states_model:
-    #     data["RESERVE_BOM({})".format(st.name)] = projector.reserves_bom[:, st]
-
-    # # add the probabilirty movements
-    # for vol_prob_res in ProbabilityVolumeResults:
-    #     data[vol_prob_res.name] = projector.probability_movements[:, vol_prob_res]
-
     df_ncf = pd.DataFrame(result_data)
     df_ncf.to_csv(outfile)
