@@ -1,33 +1,173 @@
+""" Contains the Projector objects which encapsulate the main calculations (in case of Python)
+or delegate to the C++ engine. """
+
 import logging
+from typing import Any, Optional, Union
+
 import numpy as np
+import numpy.typing as npt
+
 from pyprotolinc.assumptions.providers import AssumptionTimestepAdjustment
-from pyprotolinc import MAX_AGE
+from pyprotolinc import MAX_AGE, RunConfig
 from pyprotolinc.results import ProbabilityVolumeResults, CfNames
 from pyprotolinc.assumptions.providers import AssumptionType
+import pyprotolinc._actuarial as actuarial  # type: ignore
+from pyprotolinc.models import Model
+from pyprotolinc.portfolio import Portfolio
+from pyprotolinc.models.state_models import AbstractStateModel
+from pyprotolinc.utils import TimeAxis, TimeAxis2
+from pyprotolinc.product import AbstractProduct
+from pyprotolinc.models.model_multistate_generic import ProjectionState
+
 
 logger = logging.getLogger(__name__)
 
 
-class TimeAxis:
+class CProjector:
+    """ Encapsulate the C-kernel calls. """
 
-    def __init__(self, portfolio_date, total_num_months):
-        portfolio_year, portfolio_month = portfolio_date.year, portfolio_date.month
+    def __init__(self, run_config: RunConfig,
+                 portfolio: Portfolio,
+                 model: Model,
+                 # proj_state: Any,
+                 product: AbstractProduct,
+                 # rows_for_state_recorder: Optional[tuple[int]] = None,
+                 chunk_index: int = 1,
+                 num_chunks: int = 1) -> None:
 
-        # generate a time axis starting at the beginning of the month of the portfolio date
-        _zero_based_months_tmp = (portfolio_month - 1) + np.arange(total_num_months + 1)
-        self.months = _zero_based_months_tmp % 12 + 1
-        self.years = portfolio_year + _zero_based_months_tmp // 12
-        self.quarters = (self.months - 1) // 3 + 1
+        self.c_portfolio = actuarial.build_c_portfolio(portfolio)
+        self.time_step = actuarial.TimeStep.MONTHLY  # actuarial.TimeStep.MONTHLY  # QUARTERLY   # TODO: test if we get back a result set with this timestep
+        self.max_age = run_config.max_age
 
-    def __len__(self):
-        return len(self.months)
+        self.model = model
+        self.state_dimension: int = len(self.model.known_states)
+        self.product = product
+
+        # placeholder for the results
+        self._output_columns: list[str] = []
+        self._result: Optional[npt.NDArray[np.float64]] = None
+
+        # for logging
+        self.chunk_index = chunk_index
+        self.num_chunks = num_chunks
+
+        # construct assumption set
+        acs_be: actuarial.AssumptionSet = model.assumption_set_be  # self.build_assumption_set()
+
+        self.runner = actuarial.RunnerInterfaceWrapper(acs_be, self.c_portfolio, self.time_step, self.max_age, run_config.use_multicore, run_config.years_to_simulate)
+        self.time_axis = TimeAxis2(*self.runner.get_time_axis())
+
+        # product information, here we obtain the whole conditional payment stream upfront
+        self.cond_bom_payment_dict = self.product.get_bom_payments(self.time_axis)
+        self.cond_eom_payment_dict = self.product.get_state_transition_payments(self.time_axis)
+
+        # contractual state transitions originating from the product and the insured state
+        # rather than actuarial assumptions
+        self.contractual_state_transitions = self.product.contractual_state_transitions(self.time_axis)
+
+        # pass BOP information to C++
+        for state, payment_list in self.cond_bom_payment_dict.items():
+            for payment_type_index, payment_matrix in payment_list:
+                self.runner.add_cond_state_payment(state, payment_type_index, np.ascontiguousarray(payment_matrix))
+
+        # pass transition payments to C++
+        for (state_from, state_to), payment_list in self.cond_eom_payment_dict.items():
+            for payment_type_index, payment_matrix in payment_list:
+                self.runner.add_transition_payment(state_from, state_to, payment_type_index, np.ascontiguousarray(payment_matrix))
+
+    def run(self) -> None:
+        """ Starts the calculation run and store the results internally. """
+        self._output_columns, self._result = self.runner.run()
+
+    def get_results_dict(self) -> dict[str, Union[npt.NDArray[np.float64], npt.NDArray[np.int16]]]:
+        """ Converts the internally stored results into a dictionary and returns it. """
+        if self._result is None:
+            raise Exception("Logic Error: results must be calculated before getting them.")
+
+        # non-cash flows
+        c_output_map = {col_name: index for index, col_name in enumerate(self._output_columns) if not col_name.startswith("STATE_PAYMENT_TYPE_")}
+
+        num_rows = self._result.shape[0]
+
+        data = {
+            "YEAR": self._result[:, c_output_map["PERIOD_END_Y"]],
+            "QUARTER": (self._result[:, c_output_map["PERIOD_END_M"]] - 1) // 3 + 1,
+            "MONTH": self._result[:, c_output_map["PERIOD_END_M"]]
+        }
+
+        cf_map = {CfNames(int(col_name[19:])): index
+                  for index, col_name in enumerate(self._output_columns)
+                  if col_name.startswith("STATE_PAYMENT_TYPE_")}
+
+        output_model_map = self.model.states_model.to_std_outputs()
+
+        # cashflows
+        for cfn in CfNames:
+            mapped_col_no = cf_map.get(cfn)  # c_output_map.get(cfn.name)
+            if mapped_col_no is None:
+                data[cfn.name] = np.zeros(num_rows)
+            else:
+                data[cfn.name] = self._result[:, mapped_col_no]
+
+        # reserves -- TODO: the reserve mapping uses the names from the statet model rather than the generic output model
+        for st in self.model.states_model:
+            c_col_name = "RESERVE_BOM_{}".format(st.value)
+            output_col_name = "RESERVE_BOM({})".format(st.name)
+            mapped_col_no = c_output_map.get(c_col_name)
+            if mapped_col_no is None:
+                data[output_col_name] = np.zeros(num_rows)
+            else:
+                data[output_col_name] = self._result[:, mapped_col_no]
+
+        # add the probability movements
+        for vol_prob_res in ProbabilityVolumeResults:
+
+            # get mapping
+            mapped = output_model_map.get(vol_prob_res)
+
+            if mapped is None:
+                # add zero vector for not mapped output
+                data[vol_prob_res.name] = np.zeros(num_rows)
+            else:
+                if vol_prob_res.name.startswith("VOL_"):
+
+                    if isinstance(mapped, AbstractStateModel):
+                        state_no = mapped.value
+
+                        # TODO: decide if VOL output should be PROB_STATE or VOL_STATE
+                        c_state_name = "PROB_STATE_{}".format(state_no)
+                        data[vol_prob_res.name] = self._result[:, c_output_map[c_state_name]]
+                    else:
+                        raise Exception("VOL columns are not state transitions!")
+
+                elif vol_prob_res.name.startswith("MV_"):
+
+                    if isinstance(mapped, tuple):
+                        from_state_no = mapped[0].value
+                        to_state_no = mapped[1].value
+
+                        # TODO: decide if VOL output should be PROB_STATE or VOL_STATE
+                        c_state_name = "PROB_MVM_{}_{}".format(from_state_no, to_state_no)
+                        data[vol_prob_res.name] = self._result[:, c_output_map[c_state_name]]
+                    else:
+                        raise Exception("MVM output must be a tuple!")
+
+        return data
 
 
+# TODO: rename internal methods so that they start with "_"
 class Projector:
-    """ The projection engine. """
+    """ The PYTHON projection engine. """
 
-    def __init__(self, run_config, portfolio, model, proj_state, product,
-                 rows_for_state_recorder=None, chunk_index=1, num_chunks=1):
+    def __init__(self,
+                 run_config: RunConfig,
+                 portfolio: Portfolio,
+                 model: Model,
+                 proj_state: ProjectionState,
+                 product: AbstractProduct,
+                 rows_for_state_recorder: Optional[tuple[int]] = None,
+                 chunk_index: int = 1,
+                 num_chunks: int = 1) -> None:
 
         self.run_config = run_config
 
@@ -46,6 +186,9 @@ class Projector:
         assert self.model.states_model == product.STATES_MODEL
 
         self.sum_insured = portfolio.sum_insured
+        self.reserving_rate = portfolio.reserving_rate
+        self.years_of_birth = portfolio.years_of_birth
+        self.gender = portfolio.gender
 
         self.total_num_months = 12 * self.run_config.years_to_simulate
         self.month_count = 0
@@ -97,7 +240,7 @@ class Projector:
                                                       self.proj_state.num_records))
 
         # store ncf-results for selected policies
-        self.rows_for_payments_recorder = list(rows_for_state_recorder)
+        self.rows_for_payments_recorder: list[int] = list() if rows_for_state_recorder is None else list(rows_for_state_recorder)
         if self.rows_for_payments_recorder:
             self.payments_recorder_indexes = np.array(self.rows_for_payments_recorder, dtype=np.int32)
             self.payments_recorder = np.zeros((1 + self.total_num_months, len(self.payments_recorder_indexes), self.proj_state.num_states))
@@ -124,12 +267,17 @@ class Projector:
 
         # this container of structure (time x state x insured) will hold reserves
         # as at the begin of period
-        self.reserves_bom = None
+        self.reserves_bom: Optional[npt.NDArray[np.float64]] = None
 
-    def update_state(self, transition_ass_timestep):
+    def update_state(self, transition_ass_timestep: npt.NDArray[np.float64]) -> None:
         self.proj_state.update_state_matrix(transition_ass_timestep)
 
-    def select_applicable_base_assumptions(self, ages, genders, calendaryear):
+    def select_applicable_base_assumptions(self,
+                                           ages: npt.NDArray[np.int32],
+                                           genders: npt.NDArray[np.int32],
+                                           calendaryear: npt.NDArray[np.int32],
+                                           smokerstatus: npt.NDArray[np.int32],
+                                           yearsdisabledifdisabledatstart: npt.NDArray[np.int32]) -> None:
         """ Construct a three dimensional tensor that contains the base assumptions.
             Indexes:
               * row in the portfolio
@@ -137,25 +285,35 @@ class Projector:
               * 'to' state
         """
 
+        # print("ages, genders, calendaryear, smokerstatus, yearsdisabledifdisabledatstart")
+        # print(ages.dtype, genders.dtype, calendaryear.dtype, smokerstatus.dtype, yearsdisabledifdisabledatstart.dtype)
+
         # get BE assumptions
         for (from_state, to_state) in self.non_trivial_state_transitions_be:
             provider = self.model.rates_provider_matrix_be[from_state][to_state]
-            sel_ass = provider.get_rates(length=len(ages),
-                                         age=ages,
-                                         gender=genders,
-                                         calendaryear=calendaryear)
-            self.applicable_yearly_assumptions_be[:, from_state, to_state] = sel_ass
+            # print(from_state, to_state, provider.get_offsets())
+            if provider:
+                sel_ass = provider.get_rates(len(ages),
+                                             age=ages,
+                                             gender=genders,
+                                             calendaryear=calendaryear,
+                                             smokerstatus=smokerstatus,
+                                             yearsdisabledifdisabledatstart=yearsdisabledifdisabledatstart)
+                self.applicable_yearly_assumptions_be[:, from_state, to_state] = sel_ass
 
         # get RES assumptions
         for (from_state, to_state) in self.non_trivial_state_transitions_res:
             provider = self.model.rates_provider_matrix_res[from_state][to_state]
-            sel_ass = provider.get_rates(length=len(ages),
-                                         age=ages,
-                                         gender=genders,
-                                         calendaryear=calendaryear)
-            self.applicable_yearly_assumptions_res[:, from_state, to_state] = sel_ass
+            if provider:
+                sel_ass = provider.get_rates(len(ages),
+                                             age=ages,
+                                             gender=genders,
+                                             calendaryear=calendaryear,
+                                             smokerstatus=smokerstatus,
+                                             yearsdisabledifdisabledatstart=yearsdisabledifdisabledatstart)
+                self.applicable_yearly_assumptions_res[:, from_state, to_state] = sel_ass
 
-    def calc_payments_bom(self):
+    def calc_payments_bom(self) -> None:
         """ Calculate the payments based on the state information which should be
             as at the begin of the month. """
 
@@ -170,9 +328,7 @@ class Projector:
         if self.rows_for_payments_recorder:
             self.payments_recorder[self.month_count, :, :] = self.monthly_payment_matrix[self.payments_recorder_indexes, :]
 
-        # self.ncf_portfolio[self.month_count, :] = self.monthly_payment_matrix.sum(axis=0)
-
-    def calc_payments_eom(self):
+    def calc_payments_eom(self) -> None:
         """ Calculate the payments based on the state information which should be
             as at the end of the month (LS claims, potentially end of period reserves). """
 
@@ -198,11 +354,13 @@ class Projector:
         else:
             self.number_all_zero_results = 0
 
-    def calculate_reserves(self):
+    def calculate_reserves(self) -> npt.NDArray[np.float64]:
         """ Reserve loop: Calculated the reserves using reserving assumptions weighted by BE assumptions. """
 
-        # obtain the discount factor
-        reserving_interest = 0.0  # 0.04   # Placeholder for now!
+        # obtain the discount factor and force into column shape
+        # reserving_interest = self.reserving_rate.reshape((self.proj_state.num_records, 1))  # 0.0  # 0.04   # Placeholder for now!
+        # reserving_interest = 0.04  # Placeholder for now!
+        reserving_interest = self.reserving_rate.reshape((1, self.proj_state.num_records))
         monthly_discount_factor = (1.0 + reserving_interest)**(-1.0 / 12.0)
 
         reserves_bom_by_insured = np.zeros((len(self.time_axis),
@@ -242,11 +400,11 @@ class Projector:
             # Explanation of IDEA first: recursive calculation equation along the line of
             # reserves_bom[self.month_count, :] = CF@BOM|state=j + D * ( \sum_{states k}) p^{res, insured=i}_{j->k} (CF@EOM|state=j) + Res_bom(t+1)|state=k)
 
-            # first we determine the amounts needed bases on the transitions
+            # first we determine the amounts needed based on the transitions
             # the structure of transition_amounts is (from_state, to_state, insured), the broadcasting adds the reserve needed in the "to_state" to the payments
             transition_amounts = (cf_eom_per_state_change) + reserves_last_month_conditional   # here we reflect that reserves are not in P/L sign logic
 
-            # the transition amounts ar emultiplied with the transition probabilities
+            # the transition amounts are multiplied with the transition probabilities
             # the probabilities with time fixed have the strcuture(insured(r), from_state(f), to_state(t))
             cond_res_eom = np.einsum('rft,ftr->fr', self.transition_ass_monthly_res_with_time[self.month_count, :], transition_amounts)
 
@@ -259,7 +417,8 @@ class Projector:
 
         return reserves_bom_by_insured
 
-    def get_results_dict(self):
+    def get_results_dict(self) -> dict[str, Union[npt.NDArray[np.float64], npt.NDArray[np.int16]]]:
+        data: dict[str, Union[npt.NDArray[np.float64], npt.NDArray[np.int16]]]
         data = {
             "YEAR": self.time_axis.years,
             "QUARTER": self.time_axis.quarters,
@@ -270,8 +429,9 @@ class Projector:
             data[cfn.name] = self.ncf_portfolio[:, cfn]
 
         # add the reserves
-        for st in self.model.states_model:
-            data["RESERVE_BOM({})".format(st.name)] = self.reserves_bom[:, st]
+        if self.reserves_bom is not None:
+            for st in self.model.states_model:
+                data["RESERVE_BOM({})".format(st.name)] = self.reserves_bom[:, st]
 
         # add the probability movements
         for vol_prob_res in ProbabilityVolumeResults:
@@ -285,7 +445,10 @@ class Projector:
 
         return data
 
-    def calculate_contractual_transition_tensor(self, state_transitions_current_month):
+    def calculate_contractual_transition_tensor(self,
+                                                state_transitions_current_month: list[tuple[AbstractStateModel,
+                                                                                            AbstractStateModel,
+                                                                                            npt.NDArray[np.int32]]]) -> None:
         """ Expand the contractual transition data structure into a full transition tensor.
         """
 
@@ -306,11 +469,29 @@ class Projector:
             for st in range(num_states):
                 self.contractual_transition_ts_period[insured_selector, from_state, st] = st == to_state
 
-    def run(self):
+    def initialize_assumption_providers(self) -> None:
+        """ Call into the init hook of the assumption providers. """
+        # get BE assumptions
+        for (from_state, to_state) in self.non_trivial_state_transitions_be:
+            provider = self.model.rates_provider_matrix_be[from_state][to_state]
+            if provider is not None:
+                provider.initialize(years_of_birth=self.years_of_birth, gender=self.gender)
+
+        # get RES assumptions
+        for (from_state, to_state) in self.non_trivial_state_transitions_res:
+            provider = self.model.rates_provider_matrix_res[from_state][to_state]
+            if provider is not None:
+                provider.initialize(years_of_birth=self.years_of_birth, gender=self.gender)
+
+    def run(self) -> None:
         # projection loop over time
         _terminate = False
         _ages_last_month = -np.ones(self.proj_state.num_records)  # iniate with negative values
         _calendaryear_last_month = -np.ones(self.proj_state.num_records)
+        _years_disabled_if_at_start_last_month = np.ones(self.proj_state.num_records) * np.NaN
+
+        # initialize the assumption providers
+        self.initialize_assumption_providers()
 
         # get initial volumes
         self.probability_movements[0, :] = self.proj_state.get_monthly_probability_vol_info()
@@ -330,6 +511,7 @@ class Projector:
                 self.month_count += 1
 
                 # logger.debug("Calculating timestep {}/{}".format(yr, month))
+                # print("Timestep {}/{}".format(yr, month))
 
                 # save the accumulated state probabilities at BOM according to the BE assumptions
                 self.probability_states_with_time[self.month_count, :] = self.proj_state.probability_states[:, 1, :]
@@ -340,19 +522,23 @@ class Projector:
                 # update the yearly assumptions
 
                 # get risk factors
-                ages_months, genders = self.proj_state.get_assumption_cofactors()
+                ages_months, genders, smokerstatus, months_disabled_if_at_start = self.proj_state.get_assumption_cofactors()
                 ages = np.minimum(ages_months // 12, MAX_AGE)  # age selection depends on completed years
+                years_disabled_if_at_start = months_disabled_if_at_start // 12
+
                 calendaryear = np.ones(len(ages), dtype=np.int32) * self.time_axis.years[self.month_count]
 
                 # check if the risk factors have changed and refresh the assumptions in this case
-                assumption_update_required = not (np.array_equal(ages, _ages_last_month) and np.array_equal(calendaryear, _calendaryear_last_month))
+                assumption_update_required = not (np.array_equal(ages, _ages_last_month) and np.array_equal(calendaryear, _calendaryear_last_month)
+                                                  and np.array_equal(years_disabled_if_at_start, _years_disabled_if_at_start_last_month))
 
                 if assumption_update_required:
-                    self.select_applicable_base_assumptions(ages, genders, calendaryear)
+                    self.select_applicable_base_assumptions(ages, genders, calendaryear, smokerstatus, years_disabled_if_at_start)
 
-                    # TODO: apply assumption modifiers on policy level as found in the portfolio data
+                    # TODO: apply assumption modifiers on policy level as found in the portfolio data if any
 
                     # adjust assumptions for the length of the timestep
+                    # TODO: The adjuster could be injected here or into the model (via a Mixin?)
                     transition_ass_timestep_be = self.timestep_adjuster.adjust_simple(self.applicable_yearly_assumptions_be)
                     transition_ass_timestep_res_ts = self.timestep_adjuster_monthly.adjust_simple(self.applicable_yearly_assumptions_res)
                     # Note: for this optimization to work these variables must not be changed elsewhere
@@ -367,6 +553,7 @@ class Projector:
                 # could maybe be moved to "if"-branch above
                 _ages_last_month[:] = ages
                 _calendaryear_last_month[:] = calendaryear
+                _years_disabled_if_at_start_last_month[:] = years_disabled_if_at_start
 
                 # perform the state update for the month
                 for step in range(1, 1 + self.run_config.steps_per_month):
@@ -391,7 +578,9 @@ class Projector:
                     self.update_state(self.contractual_transition_ts_period)
 
                     # for reserves we build an updated transition matrix which we store
-                    self.transition_ass_monthly_res_with_time[self.month_count, :] = np.einsum('rij,rjk->rik', transition_ass_timestep_res_ts, self.contractual_transition_ts_period)
+                    self.transition_ass_monthly_res_with_time[self.month_count, :] = np.einsum('rij,rjk->rik',
+                                                                                               transition_ass_timestep_res_ts,
+                                                                                               self.contractual_transition_ts_period)
                 else:
 
                     # save unadjusted reserving assumptions in any case
@@ -403,7 +592,7 @@ class Projector:
 
                 self.proj_state.advance_month()
 
-                # check early termination
+                # check for early termination
                 if self.number_all_zero_results >= self.TERMINATE_AFTER_X_ZERO_MONTHS:
                     # TODO: termination flag should only be set if no future new business is still expected
                     _terminate = True
@@ -426,7 +615,7 @@ class Projector:
                     self.probability_movements[:, vol_prob_res] = fill_with_last_nonzero_value(self.probability_movements[:, vol_prob_res])
 
 
-def fill_with_last_nonzero_value(x: np.ndarray):
+def fill_with_last_nonzero_value(x: npt.NDArray[Union[np.float64, np.int32]]) -> npt.NDArray[Union[np.float64, np.int32]]:
     """ Return a vector which consists of a copy of x but with all trailing zeros
         overwritten by with the last non-zero element in the vector."""
     y = x.copy()
